# =============================================================================
# @file    test_record_from_codemeta.py
# @brief   Py.test cases for module metadata.py
# @created 2022-12-08
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/iga
# =============================================================================

import os
from   os import path
import json5
from   sidetrack import log
from   unittest.mock import patch

import iga.github                       # noqa F401
from iga.github import (
    GitHubAccount,
    GitHubRelease,
    GitHubRepo,
)


# Mocks
# .............................................................................

here      = path.dirname(path.abspath(__file__))
repo_dir  = path.join(here, 'data/fake-example/')
orcid_dir = path.join(here, 'data/orcid-examples/')


def mocked_invenio_api_available(server_url):
    return True


def mocked_invenio_token_valid(server_url):
    return True


def mocked_invenio_server_name(server_url):
    return 'TestServer'


def mocked_invenio_vocabulary(vocab):
    return [{'created': '2022-11-17T00:09:52.357057+00:00',
             'description': {'en': 'A permissive license.'},
             'icon': '',
             'id': 'apache-2.0',
             'links': {'self': 'https://data.caltechlibrary.dev/api/vocabularies/licenses/apache-2.0'},
             'props': {'osi_approved': 'y', 'scheme': 'spdx', 'url': 'http://www.apache.org/licenses/LICENSE-2.0'},
             'revision_id': 1,
             'tags': ['recommended', 'all', 'software'],
             'title': {'en': 'Apache License 2.0'},
             'type': 'licenses',
             'updated': '2022-11-17T00:09:52.368851+00:00'},
            {'created': '2022-11-17T00:09:52.356993+00:00',
             'description': {'en': ''},
             'icon': '',
             'id': 'bsd-3-clause',
             'links': {'self': 'https://data.caltechlibrary.dev/api/vocabularies/licenses/bsd-3-clause'},
             'props': {'osi_approved': 'y', 'scheme': 'spdx', 'url': 'https://opensource.org/licenses/BSD-3-Clause'},
             'revision_id': 1,
             'tags': ['all', 'software'],
             'title': {'en': 'BSD 3-Clause "New" or "Revised" License'},
             'type': 'licenses',
             'updated': '2022-11-17T00:09:52.366878+00:00'},
            {'created': '2022-11-17T00:09:52.419670+00:00',
             'description': {'en': 'A short and simple permissive license.'},
             'icon': '',
             'id': 'mit',
             'links': {'self': 'https://data.caltechlibrary.dev/api/vocabularies/licenses/mit'},
             'props': {'osi_approved': 'y', 'scheme': 'spdx', 'url': 'https://opensource.org/licenses/MIT'},
             'revision_id': 1,
             'tags': ['recommended', 'all', 'software'],
             'title': {'en': 'MIT License'},
             'type': 'licenses',
             'updated': '2022-11-17T00:09:52.430797+00:00'}]


def mocked_invenio_communities():
    return {}


def mocked_github_account(account_name):
    log(f'returing mocked GitHubAccount for {account_name}')
    with open(path.join(repo_dir, 'account.json'), 'r') as f:
        return GitHubAccount(json5.loads(f.read()))


def mocked_github_repo(account_name, repo_name):
    log(f'returing mocked GitHubRepo for {repo_name}')
    with open(path.join(repo_dir, 'repo.json'), 'r') as f:
        repo = GitHubRepo(json5.loads(f.read()))
        repo._files = mocked_github_repo_filenames(repo_name, 'faketag')
        return repo


def mocked_github_release(account_name, repo_name, tag_name):
    log(f'returing mocked GitHubRelease for {tag_name}')
    with open(path.join(repo_dir, 'release.json'), 'r') as f:
        return GitHubRelease(json5.loads(f.read()))


def mocked_github_repo_filenames(repo, tag_name):
    log('returing mocked filenames list')
    with open(path.join(repo_dir, 'filenames.json'), 'r') as f:
        return json5.loads(f.read())


def mocked_github_repo_file(repo, tag_name, filename):
    log(f'returing mocked file contents for {filename}')
    with open(path.join(repo_dir, filename), 'r') as f:
        return f.read()


def mocked_github_repo_contributors(repo):
    log('returing mocked contributors')
    return []


def mocked_github_repo_languages(repo):
    log('returing mocked languages')
    return ['fakelanguage1', 'fakelanguage2']


def mocked_orcid_data(orcid):
    log(f'returing mocked orcid data for {orcid}')
    orcid_filename = orcid + '.json'
    with open(path.join(orcid_dir, orcid_filename), 'r') as f:
        return json5.load(f)


# Tests
# .............................................................................

@patch('iga.invenio.invenio_api_available', new=mocked_invenio_api_available)
@patch('iga.invenio.invenio_token_valid', new=mocked_invenio_token_valid)
@patch('iga.invenio.invenio_server_name', new=mocked_invenio_server_name)
@patch('iga.invenio.invenio_vocabulary', new=mocked_invenio_vocabulary)
@patch('iga.invenio.invenio_communities', new=mocked_invenio_communities)
@patch('iga.github.github_account', new=mocked_github_account)
@patch('iga.github.github_release', new=mocked_github_release)
@patch('iga.github.github_repo', new=mocked_github_repo)
@patch('iga.github.github_repo_file', new=mocked_github_repo_file)
@patch('iga.github.github_repo_filenames', new=mocked_github_repo_filenames)
@patch('iga.orcid.orcid_data', new=mocked_orcid_data)
@patch.dict(os.environ, {'INVENIO_SERVER': 'data.caltechlibrary.dev',
                         'INVENIO_TOKEN': 'faketoken',
                         'GITHUB_TOKEN': 'faketoken'}, clear=True)
def test_metadata(*args):
    from iga.metadata import metadata_for_release
    record = metadata_for_release('fakeaccount', 'fakerepo', 'fakerelease', False)
    expected_json = path.join(repo_dir, 'expected-metadata.json')
    with open(expected_json, 'r') as f:
        expected = json5.loads(f.read())
    assert record == expected



# =============================================================================
# @file    test_exceptions.py
# @brief   Py.test cases for parts of exceptions.py
# @created 2023-03-02
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/iga
# =============================================================================

from iga.exceptions import IGAException, InternalError


def test_exceptions():
    try:
        raise InternalError('foo')
    except InternalError as ex:
        assert ex.args[0] == 'foo'

    try:
        raise InternalError('foo')
    except IGAException as ex:
        assert ex.args[0] == 'foo'



# =============================================================================
# @file    test_init.py
# @brief   Py.test cases for module __init__.py file
# @created 2022-12-08
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/iga
# =============================================================================

def test_version():
    """Test version import."""
    from iga import __version__
    assert __version__


def test_print_version(capsys):
    from iga import print_version
    print_version()
    captured = capsys.readouterr()
    assert 'URL' in captured.out



'''
invenio.py: Code for interacting with InvenioRDM

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2022-2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.network_utils import net, network, netloc
from   commonpy.data_utils import pluralized
import commonpy.exceptions
from   dataclasses import dataclass
from   functools import partial, cache
import json
from   sidetrack import log
import socket
import os
from   os import path

import iga
from   iga.exceptions import (
    InternalError,
    InvenioRDMError,
    RecordNotFound,
)
from   iga.github import github_asset_contents
from   iga.id_utils import normalize_invenio_rdm


# Exported data structures.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

@dataclass
class InvenioRecord():
    '''Represents the record created in InvenioRDM.

    Used internally in IGA to make certain values directly accessible and to
    make it easy to pass a record around between functions. This is not what
    InvenioRDM returns directly, but it contains values returned by InvenioRDM.
    '''
    data        : dict                  # The complete record returned by RDM.
    draft_url   : str                   # links 'self'
    review_url  : str                   # links 'review'
    publish_url : str                   # links 'publish'
    record_url  : str                   # links 'record_html' URL
    files_url   : str                   # links 'files'
    assets      : list                  # List of file names or URLs


@dataclass
class InvenioCommunity():
    '''Represents information about an Invenio community in InvenioRDM.'''
    data  : dict                       # The complete record returned by RDM.
    name  : str                        # The slug field value
    url   : str                        # The links['self_html'] field value
    id    : str                        # The id field value
    title : str                        # The metadata['title'] value


# Exported module functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def invenio_api_available(server_url):
    '''Return True if the server at server_url responds to the InvenioRDM API.'''
    # There is no 'status' or similarly simple API test endpoint for InvenioRDM,
    # so to test that the destination is actually an InvenioRDM server, the
    # approach here is to do some kind of minimum call that requires the API.
    return bool(invenio_server_name(server_url))


def invenio_token_valid(server_url):
    '''Return True if the token in env variable INVENIO_TOKEN is valid.'''
    # Testing things ahead of time is a Python anti-pattern, but in interactive
    # programs (and especially ones that run as GHAs) we don't want to get far
    # into the work before discovering we won't be able to upload the result.
    try:
        return bool(_invenio('get', endpoint='/api/users?size=1',
                             msg='confirm InvenioRDM token is valid'))
    except KeyboardInterrupt:
        raise
    except commonpy.exceptions.AuthenticationFailure:
        return False
    except Exception as ex:             # noqa PIE786
        log('exception while testing the InvenioRDM token: ' + str(ex))
        return False


@cache
def invenio_server_name(server_url):
    '''Return the name of server at server_url if it responds to API calls.'''
    server_host = netloc(server_url)
    endpoint = '/api/records?size=1'
    try:
        log(f'testing if we can reach {server_url} in 5 sec or less')
        socket.setdefaulttimeout(5)
        # If the next one can't reach the host, it'll throw an exception.
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((server_host, 443))
        # If we can reach the host, check that it responds to the API endpoint.
        if response := network('get', server_url + endpoint):
            log(f'we can reach {server_url} and it responds to {endpoint}')
            data = response.json()
            record = data.get('hits', {}).get('hits', {})[0]
            return record['metadata']['publisher']
    except KeyboardInterrupt:
        raise
    except socket.error:
        log(f'{server_url} did not respond')
        return None
    except json.decoder.JSONDecodeError:
        log(f'{server_url} exists but does not recognize the InvenioRDM API')
        return None
    except commonpy.exceptions.CommonPyException as ex:
        log(f'trying to get {server_url}{endpoint} produced an error: ' + str(ex))
        return None
    except Exception as ex:
        log(f'unexpected exception accessing {server_url}{endpoint}: ' + str(ex))
        raise
    log(f'failed to reach server {server_url}')
    return None


def invenio_get(record_id):
    '''Get the InvenioRDM record with the given identifier.'''
    log(f'asking InvenioRDM server for record {record_id}')
    record_id = normalize_invenio_rdm(record_id)
    result = _invenio('get', endpoint=f'/api/records/{record_id}',
                      msg='get record {record_id}')
    if result is None:
        # Maybe it's a draft record -- try one more time with /draft appended.
        result = _invenio('get', endpoint=f'/api/records/{record_id}/draft',
                          msg='get record {record_id}')
    if result is None:
        raise RecordNotFound(f'Could not find record with id {record_id}')
    return result


def invenio_create(metadata, parent_id=None):
    '''Create a record in the InvenioRDM server using the given metadataa.

    This only creates a draft record, and returns an InvenioRecord data class;
    it does not upload file attachments, which is something handled by a
    separate function, invenio_upload(...), or publish the record, which is
    handled by invenio_publish(...).
    '''
    # Relevant InvenioRDM documentation:
    # https://inveniordm.docs.cern.ch/reference/rest_api_drafts_records/#drafts
    # https://inveniordm.docs.cern.ch/reference/rest_api_drafts_records/#versions
    if parent_id:
        # Make sure parent actually exists. The resulting record is not needed.
        invenio_get(parent_id)

        log(f'creating a new version of record {parent_id} in InvenioRDM')
        result = _invenio('post', endpoint=f'/api/records/{parent_id}/versions',
                          msg='create new record using metadata')
        if not result:
            raise InvenioRDMError(f'Unable to create new version of {parent_id}')
        record_id = result['id']

        log(f'updating metadata in new record {record_id} in InvenioRDM')
        metadata['files'] = {'enabled': True}
        result = _invenio('put', endpoint=f'/api/records/{record_id}/draft',
                          data=metadata, msg='create new record using metadata')
        # The server will make some changes to the metadata we send.
        result['metadata'] = metadata
    else:
        log('creating record in InvenioRDM')
        result = _invenio('post', endpoint='/api/records', data=metadata,
                          msg='create new record using metadata')

    if validation_errors := result.get('errors', []):
        log(f'the server reported {pluralized("error", validation_errors, True)}:')
        for error in validation_errors:
            log(f' * in field "{error["field"]}": {error["messages"]}')
        # FIXME some should be reported to the user, but others are harmless.
    else:
        log('record created successfully with no errors')

    record = InvenioRecord(data=result,
                           draft_url=result['links']['self_html'],
                           review_url=result['links']['review'],
                           publish_url=result['links']['publish'],
                           record_url='',
                           files_url=result['links']['files'],
                           assets=[])
    log(f'new record record_html = {record.draft_url}')
    return record


def invenio_upload(record, asset, print_status):
    '''Upload a file asset to InvenioRDM and attach it to the record.

    Assets can be in one of 2 forms: a URL (a github location) or a local file.
    This will read the file into memory, then send it to the InvenioRDM server
    and commit the upload.
    '''
    # Start by reading the assets to be sure we can actually get them, *before*
    # trying to upload them to InvenioRDM.
    import humanize
    size = ''
    if asset.startswith('http'):
        filename = _filename_from_asset_url(asset)
        print_status(f' - Downloading [bold]{filename}[/] from GitHub', end='...')
        content = github_asset_contents(asset)
        print_status('done')
        size = humanize.naturalsize(len(content))
        log(f'downloaded {size} bytes of {asset}')
    else:
        filename = path.basename(asset)
        content = None
        print_status(f' - Reading [bold]{filename}[/]', end='...')
        with open(asset, 'rb') as file:
            content = file.read()
        print_status('done')
        size = humanize.naturalsize(len(content))
        log(f'read {size} bytes of file {filename}')

    # Define a helper function for the remaining steps.
    def action(op, url, dothing, **kwargs):
        log(f'doing {op} to InvenioRDM server to {dothing} {filename}')
        result = _invenio(op, url=url, msg=f'{dothing} {filename}', **kwargs)
        log(f'successfully did {op} to {dothing} {filename}')
        return result

    # Get a file upload link from the server, then using that, do a 'put' to
    # write the content followed with a 'post' to commit the new file.
    key = [{'key': filename}]
    print_status(f' - Sending [bold]{filename}[/] ({size}) to InvenioRDM', end='...')
    result = action('post', record.files_url, 'initialize upload of file', data=key)
    for entry in result.get('entries', []):
        if entry['key'] == filename:
            content_url = entry['links']['content']
            commit_url = entry['links']['commit']
            break
    else:
        log(f'server data did not contain an entry for {filename} – bailing')
        raise InternalError('Data mismatch in result from InvenioRDM')

    action('put', content_url, 'upload', data=content)
    action('post', commit_url, 'commit')
    print_status('done')


def invenio_community_send(record, community):
    '''Send the record to the InvenioRDM community for review.'''
    # cli() checks the given community exists so we don't need to do it here.
    communities = invenio_communities()
    community_id = communities[community].id

    log(f'getting the submission link for community {community} ({community_id})')
    data = {
        "receiver": {"community": community_id},
        "type": "community-submission",
    }
    result = _invenio('put', url=record.review_url, data=data,
                      msg='get community submission link from InvenioRDM')
    if submit_url := result.get('links', {}).get('actions', {}).get('submit', ''):
        data = {
            'payload': {
                'format': 'html',
                'content': 'This record is being submitted automatically using'
                f'the InvenioRDM GitHub Archiver (IGA) version {iga.__version__}',
            }
        }
        log(f'submitting the record to community {community}')
        result = _invenio('post', url=submit_url, data=data,
                          msg='submit record to community {community}')
    else:
        raise InternalError('Unexpected result in community submission step')


def invenio_publish(record):
    '''Tell the InvenioRDM server to publish the record.'''
    log('telling the server to publish the record')
    result = _invenio('post', url=record.publish_url, msg='publish record')
    record.record_url = result.get('links', {}).get('self_html', '')
    log(f'successfully published record at {record.record_url}')


@cache
def invenio_vocabulary(vocab):
    '''Return a controlled vocabulary from this server.

    The value of vocab must be one of "languages", "licenses", or
    "resourcetypes". This uses the API endpoint described in
    https://inveniordm.docs.cern.ch/reference/rest_api_vocabularies/.
    '''
    log(f'asking InvenioRDM server for vocabulary "{vocab}"')
    # At the time of this writing (2023-03-08), the "languages" vocabulary has
    # 7847 entries. We currently don't use the languages vocab in IGA, but in
    # case we ever do in the future, this code tries to get everything at once.
    endpoint = '/api/vocabularies/' + vocab + '?size=10000'
    result = _invenio('get', endpoint=endpoint, msg=f'get vocabulary {vocab}')
    if hits := result.get('hits'):
        # This is not a mistake; it really is a nested 'hits' dictionary.
        num_items = len(hits['hits'])
        log(f'received {num_items} items for vocabulary "{vocab}"')
        if num_items < hits['total']:
            log(f'warning: server actually has {hits["total"]} values available')
        return hits['hits']
    else:
        log(f'failed to get any items for vocabulary "{vocab}"')
    return []


@cache
def invenio_communities():
    '''Return a dict of communities available in this InvenioRDM server.'''
    log('asking InvenioRDM server for a list of communities')
    communities = {}
    result = _invenio('get', endpoint='/api/communities', msg='get communities list')
    for community in result.get('hits', {}).get('hits', []):
        name = community['slug']
        metadata = community['metadata']
        title = metadata.get('title', '')
        access = community['access']
        if access['visibility'] != 'public':
            log(f'community {name} is not publicly visible; skipping')
            continue
        communities[name] = InvenioCommunity(data=community,
                                             name=name,
                                             url=community['links']['self_html'],
                                             id=community['id'],
                                             title=title)
    log(f'we got {pluralized("community", communities, True)}')
    return communities


# Miscellaneous helper functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def _invenio(action, endpoint='', url='', data='', msg=''):
    '''Do HTTP action on the given endpoint with the given data & return json.

    Either (but not both) the full url, or the endpoint part, must be given as
    an argument. The data is optional, and can be a JSON dict or a binary
    string. The msg shoul describe the action being performed and will be used
    to construct log messages and an exception message if needed.
    '''
    if endpoint:
        url = os.environ.get('INVENIO_SERVER') + endpoint
    data_type = 'json' if isinstance(data, (dict, list)) else 'octet-stream'
    headers = {'Content-type': 'application/' + data_type}
    if token := os.environ.get('INVENIO_TOKEN'):
        headers['Authorization'] = 'Bearer ' + token

    if os.environ.get('IGA_RUN_MODE') == 'debug':
        d = json.dumps(data, indent=2) if data_type == 'json' else ''
        log(f'doing {action} on {url}' + (f' with payload:\n{d}' if d else ''))

    # Construct a Python partial to gather some args for calling network().
    client = None
    if action == 'put':
        # 'put' => data is being uploaded, so we need to set longer timeouts.
        import httpx
        tmout = _network_timeout(data)
        timeout = httpx.Timeout(tmout, connect=10, read=tmout, write=tmout)
        client = httpx.Client(timeout=timeout, http2=True, verify=False)
    api_call = partial(net, action, url, client=client, headers=headers)

    # Now perform the actual network api call.
    try:
        if data:
            log(f'data payload size = {len(data)}')
            if data_type == 'json':
                response, error = api_call(json=data)
            else:
                response, error = api_call(content=data)
        else:
            response, error = api_call()
        if error:
            raise error
        response_json = response.json()
        if os.environ.get('IGA_RUN_MODE') == 'debug':
            log(f'got response:\n{json.dumps(response_json, indent=2)}')
        return response_json
    except KeyboardInterrupt:
        raise
    except commonpy.exceptions.NoContent:
        log(f'got no content for {endpoint}')
        return None
    except commonpy.exceptions.CommonPyException as ex:
        # Invenio responds w/ code 400 if the token is invalid but the msg in
        # the response ("Referer checking failed") is confusing. Let's do better.
        if response.status_code == 400:
            raise InvenioRDMError(f'Failed to {msg}. The server rejected the'
                                  ' request, possibly because of an invalid'
                                  ' InvenioRDM token.')
        else:
            raise InvenioRDMError(f'Failed to {msg}. {str(ex).capitalize()}')
    except Exception:
        raise


def _network_timeout(data):
    '''Return a timeout value (in seconds) for writing the given data string.

    This uses the value in the environment variable IGA_NETWORK_TIMEOUT, if
    given. If no explicit timeout value has been given, this calculates an
    adaptive timeout value based on the size of the data, using a custom
    heuristic invented by the author based on his experiences with a
    slowish network.
    '''
    if given_timeout := os.environ.get('IGA_NETWORK_TIMEOUT'):
        # Assume that the main cli() has done sanity-checking on the value.
        timeout = int(given_timeout)
    else:
        # Calculate a value based on these principles:
        #  - Scale by the size of the data at roughly 1 min per 5 MB
        #  - Don't exceed 30 minutes (1800 sec)
        timeout = min(1800, 60*max(1, len(data)//2_000_000))
    log(f'using a network timeout of {timeout} s')
    return timeout


def _filename_from_asset_url(asset):
    '''Return a file name based on a GitHub asset URL.

    This function exists because GitHub tarball & zipball URLs end in the tag
    name, not a meaningful name, so we have to make up our own.
    '''
    parts = asset.split('/')
    if 'zipball' in asset:
        return f'{parts[-3]}_{parts[-1]}.zip'
    elif 'tarball' in asset:
        return f'{parts[-3]}_{parts[-1]}.tar.gz'
    else:
        return parts[-1]



# =============================================================================
# @file    test_name_splitting.py
# @brief   Py.test cases for parts of name_utils.py
# @created 2022-12-23
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/iga
# =============================================================================

from collections import namedtuple
from iga.name_utils import (
    split_name,
    is_person,
)

Name = namedtuple('Name', 'first last')

PARSED_NAMES = [
    ("Adam 'Atomic' Saltsman", Name(first='Adam', last='Saltsman')),
    ("Adán Miguel Sánchez Albert", Name(first='Adán Miguel Sánchez', last='Albert')),
    ("Ari", Name(first='', last='Ari')),
    ("Ching-Lan 'digdog' HUANG 黃 青嵐", Name(first='Ching-Lan', last='Huang')),
    ("Elijah Insua", Name(first='Elijah', last='Insua')),
    ("Francisco Ryan Tolmasky I", Name(first='Francisco Ryan', last='Tolmasky')),
    ("James Edward Gray II", Name(first='James Edward', last='Gray')),
    ("Jay Freeman (saurik)", Name(first='Jay', last='Freeman')),
    ("Jonathan 'Wolf' Rentzsch", Name(first='Jonathan', last='Rentzsch')),
    ("Kris Zyp", Name(first='Kris', last='Zyp')),
    ("Lim Chee Aun", Name(first='Lim Chee', last='Aun')),
    ("Matthew Weier O'Phinney", Name(first='Matthew', last="Weier O'Phinney")),
    ("Miguel de Icaza", Name(first='Miguel', last='de Icaza')),
    ("Miller Medeiros", Name(first='Miller', last='Medeiros')),
    ("Sam de Freyssinet", Name(first='Sam', last='de Freyssinet')),
    ('Andrey “A.I” Sitnik', Name(first='Andrey', last='Sitnik')),
    ('Ask Solem Hoel', Name(first='Ask Solem', last='Hoel')),
    ('Benjamin Arthur Lupton', Name(first='Benjamin Arthur', last='Lupton')),
    ('Carla Rodriguez de García', Name(first='Carla', last='Rodriguez de García')),
    ('Carlos Antonio da Silva', Name(first='Carlos Antonio', last='da Silva')),
    ('Christian Van Der Henst S.', Name(first='Christian', last='Van Der Henst S.')),
    ('Cristi Burcă', Name(first='Cristi', last='Burcă')),
    ('Dr Nic Williams', Name(first='Nic', last='Williams')),
    ('Fuji, Goro', Name(first='Fuji', last='Goro')),
    ('Georges Chreifi', Name(first='Georges', last='Chreifi')),
    ('Ian Storm Taylor', Name(first='Ian Storm', last='Taylor')),
    ('Irakli Gozalishvili', Name(first='Irakli', last='Gozalishvili')),
    ('Iuri de Silvio', Name(first='Iuri', last='de Silvio')),
    ('J. Robert Oppenheimer', Name(first='J. Robert', last='Oppenheimer')),
    ('Jasper Van der Jeugt', Name(first='Jasper', last='Van der Jeugt')),
    ('Jean-Jacques Rousseau', Name(first='Jean-Jacques', last='Rousseau')),
    ('Jose Diaz-Gonzalez', Name(first='Jose', last='Diaz-Gonzalez')),
    ('José Lorenzo Rodríguez', Name(first='José Lorenzo', last='Rodríguez')),
    ('Lukas Kahwe Smith', Name(first='Lukas Kahwe', last='Smith')),
    ('Martin Luther King, Jr.', Name(first='Martin Luther', last='King')),
    ('Miguel de Val-Borro', Name(first='Miguel', last='de Val-Borro')),
    ('Mike Hucka', Name(first='Mike', last='Hucka')),
    ('Mr.doob', Name(first='', last='Doob')),
    ('Mu-An ✌️ Chiou', Name(first='Mu-An', last='Chiou')),
    ('PatrickJS [tipe.io]', Name(first='', last='PatrickJS')),
    ('Philip (flip) Kromer', Name(first='Philip', last='Kromer')),
    ('R. Tyler Croy', Name(first='R. Tyler', last='Croy')),
    ('R.I.Pienaar', Name(first='R. I.', last='Pienaar')),
    ('Rafael França', Name(first='Rafael', last='França')),
    ('TJ Holowaychuk', Name(first='TJ', last='Holowaychuk')),
    ('TZ | 天猪', Name(first='', last='Tz')),
    ('Tokuhiro Matsuno', Name(first='Tokuhiro', last='Matsuno')),
    ('Weibin Yao(姚伟斌)', Name(first='Weibin', last='Yao')),
    ('Wladimir J. van der Laan', Name(first='Wladimir J.', last='van der Laan')),
    ('Yukihiro &quot;Matz&quot; Matsumoto', Name(first='Yukihiro', last='Matsumoto')),
    ('ara.t.howard', Name(first='Ara T.', last='Howard')),
    ('cho45', Name(first='', last='cho45')),
    ('王爵nice', Name(first='', last='Nice')),
    # ('', Name(first='', last='')),

    # Known failures:
    # ("SHIBATA Hiroshi", Name(first='Hiroshi', last='SHIBATA')),
]


def test_name_splitting():
    for original, parsed in PARSED_NAMES:
        (first, last) = split_name(original)
        assert first == parsed.first
        assert last == parsed.last



'''
ror.py: utilities for getting information from ROR

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

import commonpy.exceptions
from   commonpy.network_utils import network
from   functools import cache
import json
import os
from   sidetrack import log

from   iga.id_utils import detected_id
from   iga.exceptions import InternalError


# Internal module constants.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

_MAX_RECURSION_DEPTH = 4
'''Maximum times a lookup will follow a successor link recursively.'''


# Exported module functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def name_from_ror(rorid, recursion=0):
    '''Return the name of an organization given a ROR identifier.

    The identifier can be a pure ROR id like 05dxps055, or it can be in the
    form of a URL like https://ror.org/05dxps055.

    If the ROR record for the given id is marked as withdrawn, look to see if
    the ROR entry lists a successor. If so, do a recursive lookup.
    '''
    if not rorid or not isinstance(rorid, str):
        return ''
    rorid = detected_id(rorid) if rorid.startswith('http') else rorid
    ror_dict = ror_data(rorid)

    # A record in ROR can be marked "inactive" or "withdrawn". If inactive,
    # we still return the name because for the purposes of IGA, it does not
    # matter if an org is inactive. (We can't tell why the user used that ROR
    # id -- perhaps they're backfilling old archives, and the value may make
    # sense in the context of what the user is doing.) On the other hand, if
    # the entry is marked "withdrawn" it means the record in ROR was created
    # in error; i.e., "withdrawn" is a property of the ROR record, not the
    # organization. In that case, we try to look for a new record in ROR.
    # For more information about record statuses, see the following page:
    # https://ror.readme.io/changelog/2022-12-01-organization-status-changes

    name = ror_dict.get('name', '')
    log(f'ROR.org says organization {rorid} is named "{name}"')
    if ror_dict.get('status', '') != 'withdrawn':
        return name

    # If a record has been withdrawn, see if there's a link to an update.
    for related in ror_dict.get('relationships', []):
        if related.get('type', '').lower() == 'successor':
            new_id = related.get('id', '')
            log(f'{rorid} marked as withdrawn; its successor is ' + new_id)
            if recursion <= _MAX_RECURSION_DEPTH:
                log('doing a recursive lookup on ' + new_id)
                return name_from_ror(new_id, recursion + 1)
            else:
                log('recursion depth exceeded; not continuing ROR lookups')
                return ''
    else:
        log(f'{rorid} marked as withdrawn but ROR lists no successor')
        return ''


@cache
def ror_data(rorid):
    '''Return the data from ror.org for the given ROR id.'''
    if not rorid:
        return ''
    log(f'looking up ROR data about "{rorid}"')
    url = f'https://api.ror.org/organizations/{rorid}'
    try:
        data = network('get', url).json()
        if os.environ.get('IGA_RUN_MODE') == 'debug':
            log(f'data received for {rorid} from ror.org:\n{str(data)}')
        return data
    except KeyboardInterrupt:
        raise
    except commonpy.exceptions.NoContent:
        log(f'ROR.org returned no result for "{rorid}"')
    except commonpy.exceptions.CommonPyException as ex:
        log(f'could not get data from ROR.org for "{rorid}": ' + str(ex))
    except json.JSONDecodeError as ex:
        # This means we have to fix something.
        raise InternalError('Error trying to decode JSON from ror.org: ' + str(ex))
    except Exception:
        raise
    return ''



import iga.github
from iga.github import GitHubRelease, GitHubRepo, GitHubAccount
from os import path
import json5
from unittest.mock import patch

HERE = path.dirname(path.abspath(__file__))


def test_mocking_release(mocker):
    release_file = path.join(HERE, 'data', 'github-examples', 'with-codemeta',
                             'cds-astro', 'tutorials', 'release.json')
    with open(release_file, 'r') as f:
        release_json = json5.load(f)

    mocked_function = mocker.patch("iga.github._object_for_github")
    mocked_function.return_value = GitHubRelease(release_json)
    value = mocked_function()
    assert isinstance(value, GitHubRelease)
    assert value.id == 89397362
    assert value.author.login == 'ManonMarchand'


def test_mocking_repo(mocker):
    repo_file = path.join(HERE, 'data', 'github-examples', 'with-codemeta',
                          'cds-astro', 'tutorials', 'repo.json')
    with open(repo_file, 'r') as f:
        repo_json = json5.load(f)

    mocked_function = mocker.patch("iga.github._object_for_github")
    mocked_function.return_value = GitHubRepo(repo_json)
    value = mocked_function()
    assert isinstance(value, GitHubRepo)
    assert value.name == 'tutorials'
    assert value.owner.login == 'cds-astro'
    assert value.subscribers_count == 10


def test_mocking_account(mocker):
    account_file = path.join(HERE, 'data', 'github-examples', 'with-codemeta',
                             'datacite', 'akita', 'account.json')
    with open(account_file, 'r') as f:
        account_json = json5.load(f)

    mocked_function = mocker.patch("iga.github._object_for_github")
    mocked_function.return_value = GitHubAccount(account_json)
    value = mocked_function()
    assert isinstance(value, GitHubAccount)
    assert value.url == 'https://api.github.com/users/datacite'


def test_mocking_repo_file(mocker):
    codemeta_file = path.join(HERE, 'data', 'github-examples', 'with-codemeta',
                              'cds-astro', 'tutorials', 'codemeta.json')
    with open(codemeta_file, 'r') as f:
        codemeta_json = json5.load(f)

    mocked_function = mocker.patch("iga.github.github_repo_file")
    mocked_function.return_value = codemeta_json
    value = mocked_function('foo/repo', 'codemeta.json')
    assert isinstance(value, dict)
    assert value['codeRepository'] == 'git+https://github.com/cds-astro/tutorials'


@patch('iga.github.github_release', autospec=True, return_value = 'biff')
def test_patch_github_release(*args):
    assert iga.github.github_release('a', 'b', 'c') == 'biff'


@patch('iga.github.github_repo', autospec=True, return_value = 'bar')
def test_patch_github_repo(*args):
    assert iga.github.github_repo('a', 'b') == 'bar'


@patch('iga.github.github_release', autospec=True, return_value = 'biff')
@patch('iga.github.github_repo', autospec=True, return_value = 'bar')
def test_patch_all(*args):
    assert iga.github.github_repo('a', 'b') == 'bar'
    assert iga.github.github_release('a', 'b', 'c') == 'biff'



'''
text_utils.py: misc. utilities for doing stuff with text

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''


def cleaned_text(text):
    '''Return text that has been mildly cleaned up.

    Whitespace characters are normalized to single spaces, and period
    characters are followed by one space.
    '''
    # Get rid of embedded newlines and related characters.
    text = ' '.join(text.splitlines())
    # Normalize runs of multiple spaces to one.
    return ' '.join(text.split())


def without_html(text):
    '''Return the given text with HTML tags, if any, removed.'''
    from lxml import html
    try:
        return html.fromstring(text).text_content().strip()
    except KeyboardInterrupt:
        raise
    except Exception:                   # noqa PIE786
        return text


# Note: here's the master list of extensions for python-markdown:
# https://github.com/Python-Markdown/markdown/wiki/Third-Party-Extensions

def html_from_md(md):
    '''Convert Markdown content to HTML.'''
    from markdown import markdown
    return markdown(md, extensions=['pymdownx.superfences',
                                    'pymdownx.extra',
                                    'pymdownx.emoji',
                                    'mdx_breakless_lists',
                                    'mdx_linkify',
                                    'markdown_checklist.extension',
                                    ])



from os import path
import json5
from unittest.mock import patch

import iga.github
from iga.github import (
    GitHubRepo,
    github_account,
    github_account_repo_tag,
    github_file_url,
    valid_github_release_url)


def test_github_account():
    account = github_account('mhucka')
    assert account
    assert account.login == "mhucka"
    assert account.type == "User"

    try:
        account = github_account('55fake99zxy100')
    except iga.github.GitHubError:
        pass
    else:
        assert False


def test_github_account_repo_tag():
    url = 'https://github.com/foo/bar/releases/tag/v3.9.2'
    assert github_account_repo_tag(url) == ('foo', 'bar', 'v3.9.2')


def test_github_valid_release_url():
    assert valid_github_release_url('https://github.com/foo/bar/releases/tag/v3.9.2')
    assert not valid_github_release_url('https://github.com/foo/bar/releases/tag')


here = path.dirname(path.abspath(__file__))
json_file = 'data/github-examples/with-codemeta/fairdataihub/FAIRshare-Docs/repo.json'
with open(path.join(here, json_file), 'r') as f:
    repo_object = GitHubRepo(json5.loads(f.read()))

@patch('iga.github.github_repo', autospec=True, return_value = repo_object)
def test_repo_file_url(*args):
    repo = iga.github.github_repo('fairdataihub', 'FAIRshare-Docs')
    expected = 'https://github.com/fairdataihub/FAIRshare-Docs/blob/main/somefile'
    assert github_file_url(repo, 'somefile') == expected



# =============================================================================
# @file    test_data_utils.py
# @brief   Py.test cases for parts of data_utils.py
# @created 2023-03-02
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/iga
# =============================================================================

from typing import Generator
from iga.data_utils import (
    deduplicated,
    listified,
    similar_urls,
    normalized_url,
)


def test_deduplicated():
    assert deduplicated([]) == []
    assert deduplicated('') == []

    a = {'a': 1, 'b': 2, 'c': 3}
    assert deduplicated(a) == a
    assert deduplicated([a]) == [a]
    assert deduplicated([a, a]) == [a]

    b = {'c': 3, 'a': 1, 'b': 2}
    assert deduplicated([a, b]) == [a]

    c = {'d': {'e': {'f': 4}, 'g': 5}, 'h': 6}
    assert deduplicated([c, c]) == [c]

    d = {'d': [{'e': {'f': 4}, 'g': 5}, {'i': 7}], 'h': 6}
    assert deduplicated([d, d]) == [d]
    assert deduplicated([d, c]) == [d, c]
    assert deduplicated([d, d, a]) == [d, a]

    p = [{'person_or_org':
          {'family_name': 'Sjoberg', 'given_name': 'Daniel D.', 'type': 'personal',
           'identifiers': [{'scheme': 'orcid', 'identifier': '0000-0003-0862-2018'}]},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Bai', 'given_name': 'Xing', 'type': 'personal'},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Drill', 'given_name': 'Esther', 'type': 'personal',
           'identifiers': [{'scheme': 'orcid', 'identifier': '0000-0002-3315-4538'}]},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Flynn', 'given_name': 'Jessica', 'type': 'personal',
           'identifiers': [{'scheme': 'orcid', 'identifier': '0000-0001-8310-6684'}]},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Hannum', 'given_name': 'Margie', 'type': 'personal',
           'identifiers': [{'scheme': 'orcid', 'identifier': '0000-0002-2953-0449'}]},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Lobaugh', 'given_name': 'Stephanie', 'type': 'personal'},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Pileggi', 'given_name': 'Shannon', 'type': 'personal',
           'identifiers': [{'scheme': 'orcid', 'identifier': '0000-0002-7732-4164'}]},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Tin', 'given_name': 'Amy', 'type': 'personal',
           'identifiers': [{'scheme': 'orcid', 'identifier': '0000-0002-8005-0694'}]},
          'role': 'other'},
         {'person_or_org':
          {'family_name': 'Zapata Wainberg', 'given_name': 'Gustavo', 'type': 'personal', 'identifiers':
           [{'scheme': 'orcid', 'identifier': '0000-0002-2524-3637'}]},
          'role': 'other'}
         ]
    assert deduplicated(p) == p

    assert deduplicated(x for x in [1, 2, 3]) == [1, 2, 3]
    assert deduplicated(filter(None, [1, 2, 3])) == [1, 2, 3]
    assert deduplicated(filter(lambda x: x == 2, [1, 2, 3])) == [2]


def test_similar_urls():
    assert similar_urls('https://foo.com/bar'  , 'http://foo.com/bar')
    assert similar_urls('http://foo.com/bar'   , 'https://foo.com/bar')
    assert similar_urls('https://foo.com/bar/' , 'https://foo.com/bar')
    assert similar_urls('https://foo.com/bar'  , 'https://foo.com/bar/')
    assert similar_urls('http://foo.com/bar/'  , 'https://foo.com/bar')
    assert similar_urls('http://foo.com/bar'   , 'https://foo.com/bar/')


def test_lisitifed():
    assert listified([]) == []
    assert listified('') == []
    assert listified('a') == ['a']
    assert listified('a b') == ['a b']
    assert listified(['a']) == ['a']
    assert listified(['a', 'b']) == ['a', 'b']
    assert isinstance(listified(x for x in ['a', 'b']), Generator)


def test_normalized_url():
    assert normalized_url('git@github.com:IonicaBizau/node-git-url-parse.git') == 'https://github.com/IonicaBizau/node-git-url-parse'
    assert normalized_url('https://github.com/IonicaBizau/node-git-url-parse.git') == 'https://github.com/IonicaBizau/node-git-url-parse'
    assert normalized_url('git+ssh://git@github.com:organization/repo.git#hashbrowns') == 'https://github.com/organization/repo'
    assert normalized_url('git+https://github.com/cds-astro/tutorials') == 'https://github.com/cds-astro/tutorials'



'''
__init__.py for iga

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2022-2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

# Package metadata
# .............................................................................
#
#  ╭────────────────────── Notice ── Notice ── Notice ─────────────────────╮
#  |    The following values are automatically updated at every release    |
#  |    by the Makefile. Manual changes to these values will be lost.      |
#  ╰────────────────────── Notice ── Notice ── Notice ─────────────────────╯

__version__     = '0.0.16'
__description__ = 'InvenioRDM GitHub Archiver'
__url__         = 'https://github.com/caltechlibrary/iga'
__author__      = 'Michael Hucka'
__email__       = 'helpdesk@library.caltech.edu'
__license__     = 'https://github.com/caltechlibrary/iga/blob/main/LICENSE'


# Miscellaneous utilities.
# .............................................................................

def print_version():
    print(f'{__name__} version {__version__}')
    print(f'Authors: {__author__}')
    print(f'URL: {__url__}')
    print(f'License: {__license__}')



'''
exit_codes.py: define exit codes for program return values

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2022-2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

from aenum import Enum, MultiValue


# I adapted the clever approach posted by the author of the Python aenum
# package, Ethan Furman, to Stack Overflow on 2016-03-13 at
# https://stackoverflow.com/a/35964875/743730
# The most important bit is realizing you can define __int__().

class ExitCode(Enum):
    '''Class of exit codes that this program may return.

    The numeric value of a given code can be obtained by using int().  For
    example, int(ExitCode.success) will produce 0.
    '''

    _init_ = 'value meaning'
    _settings_ = MultiValue

    success          = 0, "success – program completed normally"
    user_interrupt   = 1, "the user interrupted the program's execution"
    bad_arg          = 2, "encountered a bad or missing value for an option"
    file_error       = 3, "encountered a problem with a file or directory"
    github_error     = 4, "encountered a problem interacting with GitHub"
    inveniordm_error = 5, "encountered a problem interacting with InvenioRDM"
    bad_token        = 6, "the personal access token was rejected"
    exception        = 7, "a miscellaneous exception or fatal error occurred"

    def __int__(self):
        return self.value



'''
doi.py: module for fetching information using DOIs

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

import commonpy.exceptions
from   commonpy.network_utils import network
from   contextlib import suppress
import functools
import json
import os
from   sidetrack import log

from   iga.id_utils import recognized_scheme
from   iga.exceptions import InternalError


# Exported module functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def doi_for_publication(pub_id, scheme=None):
    '''Given a publication id, try to get a corresponding DOI.'''
    if not pub_id:
        return ''
    scheme = recognized_scheme(pub_id) if not scheme else scheme
    if scheme == 'doi':
        return pub_id
    elif scheme == 'arxiv':
        # As of this time, not all publications in arXiv have DOIs yet, but
        # hopefully they will eventually, so we return this for all cases:
        doi = '10.48550/' + pub_id.replace(':', '.')
        log(f'converting {pub_id} into DOI {doi}')
        return doi
    elif scheme in ['pmcid', 'pmid']:
        doi = _doi_for_pubmed(pub_id, scheme)
        log(f'looked up {pub_id} and found "{doi}" as the DOI')
        return doi
    else:
        log(f'unable to get a DOI for unsupported scheme {scheme}: ' + pub_id)
    return ''


@functools.cache
def _doi_for_pubmed(pub_id, scheme):
    '''Return a DOI for a PMCID or PMID by contacting PubMed.'''
    if not pub_id:
        return ''
    log(f'looking up DOI for {scheme} {pub_id} using NCBI idconv')
    url = f'https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/?format=json&ids={pub_id}'
    try:
        data = network('get', url).json()
        if os.environ.get('IGA_RUN_MODE') == 'debug':
            log(f'data received for {scheme} "{pub_id}" from NCBI:\n{str(data)}')
        with suppress(KeyError, IndexError):
            if doi := data['records'][0].get('doi'):
                return doi
            elif errmsg := data['records'][0].get('errmsg'):
                log(f'NCBI returned an error for {pub_id}: ' + errmsg)
            else:
                log(f'did not get a DOI or error message for {pub_id} from NCBI')
    except KeyboardInterrupt:
        raise
    except commonpy.exceptions.NoContent:
        log(f'NBCI returned no result for "{pub_id}"')
    except commonpy.exceptions.CommonPyException as ex:
        log(f'could not get DOI from NCBI for "{pub_id}": ' + str(ex))
    except json.JSONDecodeError as ex:
        # This means we have to fix something.
        raise InternalError('Error trying to decode JSON from NCBI: ' + str(ex))
    except Exception:
        raise
    return ''



'''
id_utils.py: utilities for dealing with identifiers

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

from idutils import (
    detect_identifier_schemes,
    is_pmid,
    normalize_arxiv,
    normalize_doi,
    normalize_gnd,
    normalize_handle,
    normalize_isbn,
    normalize_issn,
    normalize_orcid,
    normalize_pmid,
    normalize_ror,
    normalize_urn,
)
import re


# We define our own type 'rdm' to pull out InvenioRDM identifiers from strings
# and URLs, and treat them all the same was as other identifiers. The id's are
# Douglas Crockford base 32 (https://github.com/inveniosoftware/base32-lib),
# which means 0-9 and letters except for i, l, o, and u.

rdm_regex = re.compile(r'([abcdefghjkmnpqrstvwxyz0-9]{5}-[abcdefghjkmnpqrstvwxyz0-9]{5})')
'''Matches an InvenioRDM record id.'''


def is_invenio_rdm(val):
    '''Return True if the given string is an InvenioRDM record identifier.'''
    return bool(normalize_invenio_rdm(val))


def normalize_invenio_rdm(val):
    '''Normalize an InvenioRDM record identifier.

    This accepts a URL that contains an InvenioRDM record identifier if the
    URL has one of the following forms:

       https://servername.domain/records/xxxxx-xxxxx
       https://servername.domain/uploads/xxxxx-xxxxx

    If the value is not actually an InvenioRDM identifier or a URL for a
    record containing the record identifier, this will return an empty string.
    '''
    candidate = val
    if val.startswith('http'):
        val = val.split('?')[0]         # Remove stuff like &preview=1.
        val = val.removesuffix('/draft')
        url_parts = val.split('/')
        if len(url_parts) < 2 or url_parts[-2] not in ['records', 'uploads']:
            return ''
        candidate = url_parts[-1]
    m = rdm_regex.match(candidate)
    return m.group(1) if m else ''


# The idutils regular expression for PMCID doesn't have a capturing expression
# and also, idutils lacks a "normalize_X()" for pmcid, for some reason.
PMCID_REGEX = re.compile(r"(PMC\d{6,8})", flags=re.IGNORECASE)


def contains_pmcid(val):
    '''Return True if the given value contains a PMCID.'''
    return bool(PMCID_REGEX.match(val)) if val else False


def normalize_pmcid(val):
    '''Normalize a PubMed ID.

    If the given value is not a PubMed ID, this returns an empty string.
    '''
    if not val:
        return ''
    m = PMCID_REGEX.match(val.upper())
    return m.group(1) if m else ''


def normalize_isni(val):
    '''Normalize an International Standard Name Identifier (ISNI).'''
    # FIXME not sure if anything needs to be done
    return val


def normalize_swh(val):
    '''Normalize a Software Heritage identifier (SWHID).'''
    # FIXME not sure if anything needs to be done
    return val


def normalize_ark(val):
    '''Normalize an ARK identifier.'''
    return val


def normalize_ean13(val):
    '''Normalize an EAN13 identifier.'''
    return val


def normalize_istc(val):
    '''Normalize an ISTC identifier.'''
    return val


def normalize_lsid(val):
    '''Normalize an LSID identifier.'''
    return val


def normalize_purl(val):
    '''Normalize a PURL identifier.'''
    return val


def normalize_url(val):
    '''Normalize a URL.'''
    return val


RECOGNIZED_SCHEMES = {
    'ark'    : normalize_ark,
    'arxiv'  : normalize_arxiv,
    'doi'    : normalize_doi,
    'ean13'  : normalize_ean13,
    'gnd'    : normalize_gnd,
    'handle' : normalize_handle,
    'isbn'   : normalize_isbn,
    'isni'   : normalize_isni,
    'issn'   : normalize_issn,
    'istc'   : normalize_istc,
    'lsid'   : normalize_lsid,
    'orcid'  : normalize_orcid,
    'pmcid'  : normalize_pmcid,
    'pmid'   : normalize_pmid,
    'purl'   : normalize_purl,
    'rdm'    : normalize_invenio_rdm,
    'ror'    : normalize_ror,
    'swh'    : normalize_swh,
    'urn'    : normalize_urn,
    'url'    : normalize_url,
}


def detected_id(text):
    '''Return a tuple (identifier, scheme) for an id found in text.'''
    if isinstance(text, str) and (scheme := recognized_scheme(text)):
        return RECOGNIZED_SCHEMES[scheme](text)
    return ''


def recognized_scheme(text):
    # We allow URLs that contain InvenioRDM identifiers. They're URLs & would
    # be reported as 'url' by detect_identifier_schemes, so test this case 1st.
    if is_invenio_rdm(text):
        return 'rdm'
    for scheme in detect_identifier_schemes(text):
        if scheme in RECOGNIZED_SCHEMES:
            return scheme
    else:
        # Special case not handled well by idutils. PMID's are a particular
        # PITA b/c they're integers & cause false-positives for other things.
        if is_pmid(text):
            return 'pmid'
    return None



'''
name_utils.py: utilities for dealing with people's names

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

from commonpy.data_structures import CaseFoldSet
from functools import cache
from sidetrack import log


# Internal module variables.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# The following is based on an answer posted to Stack Overflow on 2018-10-16
# by user "9999years" at https://stackoverflow.com/a/52837006/743730.
CJK_RANGES = [
    (0x4E00, 0x62FF),
    (0x6300, 0x77FF),
    (0x7800, 0x8CFF),
    (0x8D00, 0x9FCC),
    (0x3400, 0x4DB5),
    (0x20000, 0x215FF),
    (0x21600, 0x230FF),
    (0x23100, 0x245FF),
    (0x24600, 0x260FF),
    (0x26100, 0x275FF),
    (0x27600, 0x290FF),
    (0x29100, 0x2A6DF),
    (0x2A700, 0x2B734),
    (0x2B740, 0x2B81D),
    (0x2B820, 0x2CEAF),
    (0x2CEB0, 0x2EBEF),
    (0x2F800, 0x2FA1F)
]
'''List of codepoint ranges for CJK characters in Unicode. Tuples indicate
the bottom and top of the range, inclusive.'''

_NON_PERSON_ELEMENTS = {
    # Possessive expressions are almost never part of a person's name.
    "'s",
    # People who hyphenate their names don't put spaces around the hyphen.
    ' - ',
    ' – ',                              # en-dash
    ' — ',                              # em-dash
}
'''Items used as part of a filter to rule out person names.'''

_NLP = {}
'''Cache for the spaCy model so that we don't have to load it more than once.'''

_ORGANIZATIONS = {}
'''Cache for the set of well-known company names.'''

_ORGANIZATIONS_FILENAME = 'org-names.p'
'''Pickled CaseFoldSet in iga/data containing organization names.'''

_COMMON_PREFIXES = [
    '',
    'Dame',
    'Dr',
    'Fr',
    'Imām',
    'Lady',
    'Lord',
    'Messrs',
    'Miss',
    'Mr',
    'Mrs',
    'Ms',
    'Mx',
    'Pr',
    'Prof',
    'Professor',
    'Rabbi',
    'Revd',
    'Roshi',
    'Sensei',
    'Sir',
]


# Exported module functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# https://stackoverflow.com/questions/54334304/spacy-cant-find-model-en-core-web-sm-on-windows-10-and-python-3-5-3-anacon
# import spacy.cli
# spacy.cli.download("en_core_web_trf")
# spacy.cli.download("zh_core_web_trf")

@cache
def is_person(name):
    '''Try to guess whether the given string is a person's name.

    In the most general case, it is impossible to reliably determine whether a
    given string of characters represents the name of a person and not, say,
    the name of a company or product. This function makes a best-effort guess
    using heuristics in combination with methods from natural language
    processing (NLP), but it will sometimes make mistakes.

    This function is designed with the assumption that its input is already
    expected to be a name, and it only needs to make the determination of
    whether it is the name of a person or not.
    '''
    if not name:
        return False

    # A string like "Joe's Foobar" is not a name.
    if any(item in name for item in _NON_PERSON_ELEMENTS):
        log(f'{name} contains non-person elements => not a person')
        return False

    # If the input contains a mix of Latin and CJK characters, it's likely to
    # contain both an English and CJK version of the same thing. (E.g.,
    # someone might write their name in Chinese in parentheses after an English
    # version.) If that's the case, we use only the non-CJK part if there is
    # one. If there are no CJK characters, we always use the cleaned version.
    charset = 'cjk' if contains_cjk(name) else 'default'
    log(f'{name} uses the {charset} charset')
    cleaned = _cleaned_name(name)
    if charset == 'default' or len(cleaned) > 0:
        log(f'using cleaned version of name: {cleaned}')
        name = cleaned
    if not name:
        # Cleaning removed everything.
        log(f'cleaning {name} resulted in an empty string => returning False')
        return False

    # Reject names whose every part is a number.
    import re
    name_tokens = re.split(r'[-\s:]+', name)
    if all(token.isdigit() for token in name_tokens):
        log(f'{name} consists of just numbers => not a person')
        return False

    # The ML-based NER systems sometimes mislabel company names, so we start
    # by checking against a list of known organization names.
    global _ORGANIZATIONS
    if not _ORGANIZATIONS:
        _load_organizations()
    if name in _ORGANIZATIONS:
        log(f'recognized {name} as a known organization => not a person')
        return False

    # Delay loading the ML systems until needed because they take long to load.
    # We check we succeded in loading it & fall back to other things if not.
    global _NLP
    if charset not in _NLP:
        _load_spacy(charset)
    spacy_loaded = (charset in _NLP)

    def person_according_to_spacy(name):
        try:
            parsed = _NLP[charset](name)
            if parsed.ents:
                entity_type = parsed.ents[0].label_
                log(f'spaCy [{charset}] entity type for {name}: {entity_type}')
                return (entity_type == 'PERSON')
            else:
                log(f'spaCy did not return entity labels for {name}')
                # Note the return is None, not False, so caller can test.
                return None
        except KeyboardInterrupt:
            raise
        except Exception as ex:             # noqa: PIE786
            log('unable to use spaCy due to error: ' + str(ex))
        return False

    def person_according_to_pp(name):
        import probablepeople as pp
        try:
            from_pp = pp.tag(name)
            log(f'entity type according to PP for {name}: {from_pp[1]}')
            return (from_pp[1] == 'Person')
        except KeyboardInterrupt:
            raise
        except Exception:                 # noqa: PIE786
            # PP sometimes fails with an exception.
            log(f'probablepeople failed for {name}')
        return False

    # Check that we successfully loaded spaCy & fall back to PP if we didn't.
    decision = None
    if spacy_loaded:
        decision = person_according_to_spacy(name)
    if decision is None:
        # spaCy does the best job, but sometimes it doesn't produce any tags.
        # We fall back to PP in that case.
        decision = person_according_to_pp(name)
    log(f'final decision: is_person({name}) = {decision}')
    return decision


@cache
def split_name(name):
    '''Split a name into given name & surname.

    Note: trying to split names automatically into parts is not only impossible
    in general -- it's also undesirable: not all cultures use names with only
    first and/or last name components, or just one last name, or in the same
    order as Western names, etc. However, we have no choice for IGA because
    InvenioRDM's record format only supports split first & last names.
    '''
    # The approach taken here is roughly:
    #  1) strip non-Latin characters and non-alpha characters from the string
    #  2) if the result is only a single word, treat it as a surname only
    #  3) invoke a machine learning-based name parser (ProbablePeople) to do
    #     a best-effort attempt at splitting the name into given + surname
    #  4) PP is prone to failures on some names, so if it fails, fall back to
    #     using a different name parser (nameparser) that doesn't split names
    #     quite as correctly as PP, but is better than nothing.

    log('splitting name ' + name)
    name = _cleaned_name(name)
    if len(name.split(' ')) == 1:
        # Only one word in the name. Either it is really a single name (e.g.,
        # in cultures where people have single names) or someone is being cute.
        log('treating single name as the family name')
        given = ''
        surname = name
    else:
        try:
            log('trying to split name using probablepeople')
            import probablepeople as pp

            # PP gets better results if you DON'T supply the 'type' parameter.
            # (I don't know why.) Try it that way first, let it guess the type,
            # and if it guesses wrong, try it again but with the type that time.
            from_pp = pp.tag(name)
            log(f'probablepeople initial result: {str(from_pp)}')
            guessed_type = from_pp[1]
            guessed_prefix = from_pp[0].get('PrefixOther', '')
            has_bad_prefix = guessed_prefix not in _COMMON_PREFIXES
            has_corp_key = any(key.startswith('Corp') for key in from_pp[0].keys())
            if guessed_type != 'Person' or has_bad_prefix or has_corp_key:
                from_pp = pp.tag(name, type='person')
                log(f'probablepeople 2nd result: {str(from_pp)}')
                guessed_prefix = from_pp[0].get('PrefixOther', '')
                if guessed_prefix not in _COMMON_PREFIXES:
                    # This is a sign something is still wrong. Give up on PP.
                    raise Exception
            parsed = from_pp[0]

            # If it found initials instead of full names, use those.
            if parsed.get('FirstInitial'):
                given = parsed.get('FirstInitial')
            else:
                # Get rid of periods at the end, in case someone got cute.
                given = parsed.get('GivenName', '').rstrip('.')

            # InvenioRDM wants middle names to be part of the first/given names:
            if middle := (parsed.get('MiddleInitial') or parsed.get('MiddleName')):
                given += (' ' + middle)

            # If we only have a last initial, then that's all we can use.
            if parsed.get('LastInitial') and not parsed.get('Surname'):
                surname = parsed.get('LastInitial', '').title()
            else:
                surname = parsed.get('Surname', '')

            # If we didn't find a given name but the surname has a space, use
            # it as a split point & take all but last word as the given name.
            if not given and ' ' in surname:
                split = surname.split(' ')
                surname = split.pop()
                given = ' '.join(split)
        except KeyboardInterrupt:
            raise
        except Exception:                   # noqa: PIE786
            log(f'switching to nameparser b/c probablepeople failed on "{name}"')
            from nameparser import HumanName
            parsed = HumanName(name)
            # (Noted above) InvenioRDM includes middle name as part of 1st name:
            given = parsed.first + ' ' + parsed.middle
            surname = parsed.last

    given = _first_letters_upcased(given)
    if _plain_word(surname):
        surname = surname.title()
    surname = surname.strip()

    log(f'final name splitting result: ({given}, {surname})')
    return (given, surname)


def flattened_name(name):
    '''Return name as a string even if it's a list and not a simple string.'''
    return ' '.join(part for part in name) if isinstance(name, list) else name


def contains_cjk(text):
    '''Return True if text contains any character in the CJK character sets.'''
    def is_cjk(char):
        char = ord(char)
        return any(char >= bottom and char <= top for bottom, top in CJK_RANGES)

    return any(map(is_cjk, text))


# Miscellaneous helper functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def _cleaned_name(name):
    import re
    import demoji
    from iga.text_utils import without_html
    # Remove any HTML tags there might be left.
    name = without_html(name)
    # Remove parenthetical text like "Somedude [somedomain.io]".
    name = re.sub(r"\(.*?\)|\[.*?\]", "", name)
    # Remove CJK characters because the name parsers can't handle them.
    # This regex is from https://stackoverflow.com/a/2718268/743730
    name = re.sub(u'[⺀-⺙⺛-⻳⼀-⿕々〇〡-〩〸-〺〻㐀-䶵一-鿃豈-鶴侮-頻並-龎]', '', name)
    # Remove miscellaneous weird characters if there are any.
    name = demoji.replace(name)
    name = re.sub(r'[~`!@#$%^&*_+=?<>(){}|[\]¡¿]', '', name)
    # Replace typographical quotes with regular quotes, for PP's benefit.
    name = re.sub(r'[“”‘’]', '"', name)
    # Make sure periods are followed by spaces.
    name = name.replace('.', '. ')
    # Normalize runs of multiple spaces to one.
    name = re.sub(r' +', ' ', name)
    return name.strip()                 # noqa PIE781


def _plain_word(name):
    return (' ' not in name
            and not any(str.isdigit(c) for c in name)
            and (all(str.isupper(c) for c in name)
                 or not any(str.isupper(c) for c in name[1:])))


def _first_letters_upcased(name):
    # Python's .title() will downcase the letters after the 1st letter, which
    # is undesired behavior for the situation where we need this.
    return ' '.join(word[0].upper() + word[1:] for word in name.split())


def _load_organizations():
    global _ORGANIZATIONS
    from os.path import dirname, abspath, join
    import pickle
    here = dirname(abspath(__file__))
    org_file = join(here, f'data/{_ORGANIZATIONS_FILENAME}')
    log(f'loading {org_file} – this may take some time')
    with open(org_file, 'rb') as f:
        _ORGANIZATIONS = pickle.load(f)


def _load_spacy(charset):
    global _NLP
    import spacy
    model = 'zh_core_web_trf' if charset == 'cjk' else 'en_core_web_trf'
    try:
        log(f'try to load spaCy model {model} – this may some time')
        _NLP[charset] = spacy.load(model)
    except KeyboardInterrupt:
        raise
    except OSError:
        log(f'spaCy {model} not yet installed, so doing one-time download')
        if _successful_spacy_download(model):
            log('spaCy model has been successfully downloaded')
            _NLP[charset] = spacy.load(model)
        else:
            log(f'unable to get download {model} – spaCy will not be usable')
    else:
        log(f'finished loading spaCy pipeline for {charset} charset')


def _successful_spacy_download(model):
    from spacy.cli import download
    from subprocess import SubprocessError
    try:
        download(model)
    except KeyboardInterrupt:
        raise
    except (FileNotFoundError, SubprocessError) as ex:
        log(f'error trying to download spaCy model {model}: {str(ex)}')
    else:
        return True
    return False



'''
orcid.py: simple functions for interfacing to ORCID

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

import commonpy.exceptions
from   commonpy.network_utils import network
from   functools import cache
import json
import os
from   sidetrack import log

from   iga.id_utils import detected_id
from   iga.exceptions import InternalError
from   iga.name_utils import split_name


# Exported module functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def name_from_orcid(orcid):
    '''Return a (first name, family name) tuple for the given ORCID id.

    The identifier can be a pure ORCID id like 0000-0001-9105-5960, or it can
    be in the form of a URL like https://orcid.org/0000-0001-9105-5960.

    ORCID records often provide split family/given names, though not always.
    If a given user's name is not available in a split form but a name is
    available in the form of an undifferentiated name, attempt to split the
    name using methods in IGA's name_utils module.

    Records in ORCID can be deprecated or deactivated, but the public JSON
    record data does not contain successor or replacement record information.
    Return empty strings in that case.
    '''
    if not orcid or not isinstance(orcid, str):
        return ('', '')
    orcid = detected_id(orcid) if orcid.startswith('http') else orcid
    orcid_dict = orcid_data(orcid)

    # The public record JSON has no status field. If a record is deprecated or
    # deactivated, there's no explicit indicator AND we have no way to find the
    # replacement even though visiting orcid.org shows it. We're stuck w/ this:
    if 'family name deactivated' in orcid_dict.get('displayName', '').lower():
        log(f'{orcid} is a deactivated record')
        return ('', '')

    given = ''
    family = ''
    if names := orcid_dict.get('names', {}):
        # The creditName value is preferred for our uses (giving credit for
        # work) and may be more complete than the givenName & familyName field
        # values, but it's a single string. Splitting names is error-prone, so
        # we compromise: use our name splitter & use only the given name from
        # the result b/c it's the part we suspect will be preferred over the
        # givenNames field value anyway, and always use the familyName value.
        if (cn := names.get('creditName', {})) and (value := cn.get('value', '')):
            log('record has a creditName value: ' + value)
            given, family = split_name(value)
            if family_dict := names.get('familyName', {}):
                family = family_dict.get('value')
            log('final composite name: ' + given + ' ' + family)
        else:
            log('record does not have a creditName value')
            if given_dict := names.get('givenNames', {}):
                given = given_dict.get('value', '')
            if family_dict := names.get('familyName', {}):
                family = family_dict.get('value')
            log('values from given and family names: ' + given + ' ' + family)

        # In some societies, people only have a single name. If we get that
        # much from the name field, we assume we're done.
        if family:
            log(f'returning "({given}, {family})" for {orcid}')
            return (given, family)

    # The "names" object is empty for some reason. See if there's another name.
    if other := orcid_dict.get('otherNames', {}):
        log(f'"names" is empty in record for {orcid} but "otherNames" exists')
        # This dict really has a nested dict with the same name. Don't ask me.
        for item in other.get('otherNames', []):
            name = item.get('content', '') or item.get('sourceName', '')
            if name:
                log('found otherNames dict in ORCID data with name ' + name)
                given, family = split_name(name)
                break
            else:
                log('found otherNames dict in ORCID data without a name')
        else:
            log('failed to find inner otherNames dict in ORCID otherNames')
    log(f'returning "({given}, {family})" for {orcid}')
    return (given, family)


@cache
def orcid_data(orcid):
    '''Return the data from orcid.org for the given orcid id.'''
    if not orcid:
        return ''
    log(f'looking up ORCID data for "{orcid}"')
    url = f'https://orcid.org/{orcid}/public-record.json'
    try:
        data = network('get', url).json()
        if os.environ.get('IGA_RUN_MODE') == 'debug':
            log(f'data received for {orcid} from orcid.org:\n{str(data)}')
        return data
    except KeyboardInterrupt:
        raise
    except commonpy.exceptions.NoContent:
        log(f'orcid.org returned no result for "{orcid}"')
    except commonpy.exceptions.CommonPyException as ex:
        log(f'could not get data from orcid.org for "{orcid}": ' + str(ex))
    except json.JSONDecodeError as ex:
        # This means we have to fix something.
        raise InternalError('Error trying to decode JSON from orcid.org: ' + str(ex))
    return ''



# =============================================================================
# @file    test_reference.py
# @brief   Py.test cases for parts of reference.py
# @created 2023-03-02
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/iga
# =============================================================================

from iga.reference import (
    reference,
    reference_from_bibtex,
    reference_from_doi,
)


_BIBTEX_TESTS = [
    ('''@inproceedings{Myers2017briefb,
    title = {A brief history of {COMBINE}},
    doi = {10.1109/WSC.2017.8247840},
    booktitle = {2017 {Winter} {Simulation} {Conference} ({WSC})},
    author = {Myers, C. J. and Bader, G. and Gleeson, P. and Golebiewski, M. and Hucka, M. and Nov{\\`e}re, N. Le and Nickerson, D. P. and Schreiber, F. and Waltemath, D.},
    month = dec,
    year = {2017},
    pages = {884--895},
    }''', '''Myers, C. J., Bader, G., Gleeson, P., Golebiewski, M., Hucka, M., Novère, N. L., Nickerson, D. P., Schreiber, F., & Waltemath, D. (2017, December). A brief history of COMBINE. 2017 Winter Simulation Conference (WSC) (pp. 884–895). doi:10.1109/WSC.2017.8247840'''),

    ('''@article{bornstein2008libsbml,
    title={{L}ib{SBML}: an {API} library for {SBML}},
    author={Bornstein, Benjamin J and Keating, Sarah M and Jouraku, Akiya and Hucka, Michael},
    journal={Bioinformatics},
    volume={24},
    number={6},
    pages={880--881},
    year={2008},
    doi = {10.1093/bioinformatics/btn051},
    publisher={Oxford University Press}
    }''', '''Bornstein, B. J., Keating, S. M., Jouraku, A., & Hucka, M. (2008). LibSBML: an API library for SBML. Bioinformatics, 24(6), 880–881. doi:10.1093/bioinformatics/btn051'''),

    ('''@article{courtot2011controlled,
    author = {Courtot, M\\'{e}lanie  and Juty, Nick J. and Kn\\"{u}pfer, Christian  and Waltemath, Dagmar  and Zhukova, Anna  and Dr\\"{a}ger, Andreas  and Dumontier, Michel  and Finney, Andrew  and Golebiewski, Martin  and Hastings, Janna  and Hoops, Stefan  and Keating, Sarah  and Kell, Douglas B.  and Kerrien, Samuel  and Lawson, James  and Lister, Allyson  and Lu, James  and Machne, Rainer  and Mendes, Pedro  and Pocock, Matthew  and Rodriguez, Nicolas  and Villeger, Alice J. and Wilkinson, Darren J.  and Wimalaratne, Sarala  and Laibe, Camille  and Hucka, Michael  and Le Nov\\`{e}re, Nicolas},
    title = {Controlled vocabularies and semantics in systems biology},
    journal = {Molecular Systems Biology},
    volume = {7},
    number = {543},
    pages = {1},
    month = oct,
    year = {2011},
    doi = {10.1038/msb.2011.77},
    }''', '''Courtot, M., Juty, N. J., Knüpfer, C., Waltemath, D., Zhukova, A., Dräger, A., Dumontier, M., Finney, A., Golebiewski, M., Hastings, J., Hoops, S., Keating, S., Kell, D. B., Kerrien, S., Lawson, J., Lister, A., Lu, J., Machne, R., Mendes, P., Pocock, M., et al. (2011, October). Controlled vocabularies and semantics in systems biology. Molecular Systems Biology, 7(543), 1. doi:10.1038/msb.2011.77'''),

    ('''@article{Draeger2009b,
    author = {Dr\\"ager, Andreas and Planatscher, Hannes and Wouamba, Dieudonn{\\'e}
    Motsou and Schr\\"oder, Adrian and Hucka, Michael and Endler, Lukas
    and Golebiewski, Martin and M\\"uller, Wolfgang and Zell, Andreas},
    title = {{SBML2LaTeX: Conversion of SBML files into human-readable reports}},
    journal = {Bioinformatics},
    year = {2009},
    volume = {25},
    pages = {1455--1456},
    number = {11},
    month = apr,
    doi = {10.1093/bioinformatics/btp170},
    pdf = {http://bioinformatics.oxfordjournals.org/cgi/reprint/25/11/1455.pdf},
    }''', '''Dräger, A., Planatscher, H., Wouamba, D. M., Schröder, A., Hucka, M., Endler, L., Golebiewski, M., Müller, W., & Zell, A. (2009, April). SBML2LaTeX: Conversion of SBML files into human-readable reports. Bioinformatics, 25(11), 1455–1456. doi:10.1093/bioinformatics/btp170'''),

    ('''@book{Klipp:SystemsBiologyATextbook:2011,
    title={Systems biology: a textbook},
    author={Klipp, Edda and Liebermeister, Wolfram and Wierling, Christoph and Kowald, Axel and Herwig, Ralf},
    year={2016},
    publisher={John Wiley \\& Sons}
    }''', '''Klipp, E., Liebermeister, W., Wierling, C., Kowald, A., & Herwig, R. (2016). Systems biology: a textbook. John Wiley & Sons.'''),

    ('''@incollection{chaouiya2012logical,
    title={Logical modelling of gene regulatory networks with {GINsim}},
    author={Chaouiya, Claudine and Naldi, Aur\\'{e}lien and Thieffry, Denis},
    booktitle={Bacterial Molecular Networks},
    pages={463--479},
    year={2012},
    publisher={Springer}
    }''', '''Chaouiya, C., Naldi, A., & Thieffry, D. (2012). Logical modelling of gene regulatory networks with GINsim. Bacterial Molecular Networks (pp. 463–479). Springer.'''),

    ('''@techreport{Watanabe2018dynamic,
    author = {Watanabe, Leandro H. and K{\\"o}nig, Matthias and Myers, Chris J.},
    title = {Dynamic Flux Balance Analysis Models in SBML},
    year = {2018},
    doi = {10.1101/245076},
    institution = {Cold Spring Harbor Laboratory},
    URL = {https://www.biorxiv.org/content/early/2018/01/08/245076},
    eprint = {https://www.biorxiv.org/content/early/2018/01/08/245076.full.pdf},
    journal = {bioRxiv}
    }''', '''Watanabe, L. H., König, M., & Myers, C. J. (2018). Dynamic Flux Balance Analysis Models in SBML. Cold Spring Harbor Laboratory.'''),
]


def test_reference_from_bibtex():
    for bibtex, formatted in _BIBTEX_TESTS:
        assert reference_from_bibtex(bibtex) == formatted


def test_reference():
    assert reference('PMC4908318') == 'Gómez, H. F., Hucka, M., Keating, S. M., Nudelman, G., Iber, D., & Sealfon, S. C. (2016). MOCCASIN: converting MATLAB ODE models to SBML. Bioinformatics, 32(12), 1905–1906. https://doi.org/10.1093/bioinformatics/btw056'
    assert reference('978-1848162204') == 'Bolouri, H. (2008). Computational Modeling Of Gene Regulatory Networks - A Primer. Imperial College Press.'



'''
json_utils.py: utilities for working with JSON

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

import dirtyjson
from   sidetrack import log


# Internal module constants.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

_MAX_RECURSION_DEPTH = 4
'''Maximum times we will retry parsing the json content.'''


# Exported module functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def partial_json(content, skip_line=None, recursion=0):
    '''Parse content as JSON, skipping invalid lines, and return a dict.

    The approach is simple-minded: try to parse the file, and if an error
    occurs, skip the line where the error was first detected. Sometimes an
    error affects online one line and the rest of the JSON content may be
    usable. Other times, skipping a line will result in completely unparseable
    JSON, so the result of using this function may be an empty JSON.
    '''
    if skip_line:
        log(f'skipping line {skip_line} in JSON content and retrying')
        # This is an embarrassingly simple-minded approach. Maybe someday we
        # can make something better and more performant, but for now ...
        content = content.split('\n')
        content = content[:skip_line - 2] + content[skip_line - 1:]
        content = '\n'.join(content)
    try:
        return dict(dirtyjson.loads(content))
    except KeyboardInterrupt:
        raise
    except dirtyjson.error.Error as ex:
        # Do we have a line number?
        log('error parsing JSON: ' + str(ex))
        if line := getattr(ex, 'lineno'):
            if recursion <= _MAX_RECURSION_DEPTH:
                return partial_json(content, line, recursion + 1)
            else:
                log('exceeded recursion depth and admitting failure')
                return {}
    return {}



'''
reference.py: create a formatted reference for a given publication.

This file is part of https://github.com/caltechlibrary/iga/.

Copyright (c) 2023 by the California Institute of Technology.  This code
is open-source software released under a BSD-type license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.network_utils import network
import commonpy.exceptions
import json
from   sidetrack import log

from iga.doi import doi_for_publication
from iga.exceptions import InternalError
from iga.id_utils import recognized_scheme
from iga.text_utils import without_html


# Internal variables for this module.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

_CACHE = {}
'''Internal cache used to store results of some operations across calls.'''


# Exported constants.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

RECOGNIZED_REFERENCE_SCHEMES = ['arxiv', 'doi', 'isbn', 'pmcid', 'pmid']


# Exported module functions.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def reference(pub_id):
    '''Given an id (e.g., a DOI), return a formatted APA-style reference.'''
    scheme = recognized_scheme(pub_id)
    log(f'reference {pub_id} has scheme {scheme}')

    formatted_reference = ''
    if scheme == 'isbn':
        # ISBNs cannot be converted to DOIs & have to be handled separately.
        import isbnlib
        from   isbnlib.registry import bibformatters as isbn_bibformatters
        if isbn_metadata := isbnlib.meta(pub_id):
            log(f'got metadata for ISBN {pub_id}: ' + str(isbn_metadata))
            bibtex_string = isbn_bibformatters['bibtex'](isbn_metadata)
            formatted_reference = reference_from_bibtex(bibtex_string)
        else:
            # Sometimes ISBN can't be found. Not clear what can be done here.
            log(f'could not find data for ISBN {pub_id}')
    elif scheme in RECOGNIZED_REFERENCE_SCHEMES:
        # For everything other than ISBN, convert whatever ID we have to a
        # DOI, then use Crossref to get a reference as text in APA format.
        if doi := doi_for_publication(pub_id, scheme):
            formatted_reference = reference_from_doi(doi)
        else:
            log(f'could not get a DOI for {pub_id}')
    else:
        log(f'cannot convert a {scheme} type id into a formatted reference')
        return ''

    log(f'returning formatted result for {pub_id}: ' + formatted_reference)
    return formatted_reference


def reference_from_doi(doi):
    '''Given a DOI, return an APA-style formatted reference.

    This function takes advantage of Crossref's network service for generating
    reference strings in APA format. It cleans the result of HTML tags.
    '''
    global _CACHE
    cache_key = doi + '-reference'
    if cache_key in _CACHE:
        cached = _CACHE[cache_key]
        log(f'returning cached reference for {doi}: ' + cached)
        return cached

    log(f'asking DOI.org for formatted reference for {doi}')
    doi_url = 'https://doi.org/' + doi
    headers = {'accept': 'text/x-bibliography; style=apa'}
    try:
        response = network('get', doi_url, headers=headers)
        log('received response from Crossref:\n' + response.text)
        text = without_html(response.text)
        _CACHE[cache_key] = text
        return text
    except KeyboardInterrupt:
        raise
    except commonpy.exceptions.NoContent:
        log(f'CrossRef returned no result for "{doi}"')
    except commonpy.exceptions.CommonPyException as ex:
        log(f'could not get data from CrossRef for "{doi}": ' + str(ex))
    except json.JSONDecodeError as ex:
        # This means we have to fix something.
        raise InternalError('Error trying to decode JSON from CrossRef: ' + str(ex))
    except Exception:
        raise
    return ''


def reference_from_bibtex(bibtex_string):
    '''Give a string containing a BibTeX entry, return an APA-style reference.'''
    from pybtex.plugin import find_plugin
    from pybtex.database import parse_string
    import latexcodec                   # noqa F401

    # Cache the results of these plugin lookups for greater efficiency.
    global _CACHE
    if 'apa_style' in _CACHE:
        apa_style = _CACHE['apa_style']
        plain_text = _CACHE['plain_text']
    else:
        log('loading pybtex plugins')

        # Currently have to use our own patched version of pybtex-apa7-style.
        # apa_style = find_plugin('pybtex.style.formatting', 'apa7')()
        from .vendor.pybtex_apa7_style.formatting.apa import APAStyle
        apa_style = APAStyle()
        plain_text = find_plugin('pybtex.backends', 'text')()
        _CACHE['apa_style'] = apa_style
        _CACHE['plain_text'] = plain_text

    bib_data = parse_string(bibtex_string, 'bibtex')
    formatted_bib = apa_style.format_bibliography(bib_data)

    # formatted_bib is an iterable object, but not an iterator. We know there's
    # only one, so just do a next() after creating an iterator out of it.
    formatted_item = next(iter(formatted_bib))
    return formatted_item.text.render(plain_text)



# =============================================================================
# @file    test_text_utils.py
# @brief   Py.test cases for parts of text_utils.py
# @created 2023-03-02
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/iga
# =============================================================================

from typing import Generator
from iga.text_utils import (
    cleaned_text,
    without_html,
)


def test_cleaned_text():
    assert cleaned_text('') == ''
    assert cleaned_text('a  b') == 'a b'
    assert cleaned_text('a\nb') == 'a b'
    assert cleaned_text('a\t b') == 'a b'
    assert cleaned_text('some text.') == 'some text.'


def test_without_html():
    assert without_html('') == ''
    assert without_html('a') == 'a'
    assert without_html('this has no html') == 'this has no html'
    assert without_html('foo <i>bar</i>') == 'foo bar'
    assert without_html('Sjoberg, D., D., Whiting, K., Curry, M., Lavery, J., A., & Larmarange, J. (2021). Reproducible Summary Tables with the gtsummary Package. The R Journal, 13(1), 570. https://doi.org/10.32614/rj-2021-053\n') == 'Sjoberg, D., D., Whiting, K., Curry, M., Lavery, J., A., & Larmarange, J. (2021). Reproducible Summary Tables with the gtsummary Package. The R Journal, 13(1), 570. https://doi.org/10.32614/rj-2021-053'
